﻿<%@ WebService Language="C#" Class="SPMetaData" %>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Services;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Utilities;
using System.Text;
using System.Configuration;
using System.Text.RegularExpressions;
using System.Xml;


/// <summary>
/// Summary description for SPMetaData
/// </summary>
//[WebService(Namespace = "http://tempuri.org/")]
[WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
// To allow this Web Service to be called from script, using ASP.NET AJAX, uncomment the following line. 
// [System.Web.Script.Services.ScriptService]
public class SPMetaData : System.Web.Services.WebService
{
    private bool runWithElevatedPrivileges = false;
    private static bool verbose = false;
    private static string logFilePath = string.Empty;
    private static bool fixClaimsUsername = false;
    private static int itemsPerPage;
    private static int pagesPerBatch;
    private static string currentPagingInfo = string.Empty;
    XmlDocument result = new XmlDocument();
    
    public SPMetaData () {

        //Uncomment the following line if using designed components 
        InitializeComponent(); 
    }

    #region WEBMETHODS

    //[WebMethod(Description = "This method gets a string of delimited SharePoint Lists URLs, a string of delimited metadata fields and a modification date in this format: mm/dd/yyyy hh:mm:ss AM or PM. It returns an xml of the SharePoint items IDs that were modified since the modification date with their field values.")]
    [WebMethod]
    public XmlDocument GetItemsFieldsDataByListURL(string lists, string fields, string lastModified, int nextItemID)
    {
        try
        {
            //StringBuilder res = new StringBuilder();
            DateTime _lastModified;

            if (!DateTime.TryParse(lastModified, out _lastModified))
            {
                _lastModified = new DateTime(2000, 1, 1);
            }


            if (runWithElevatedPrivileges)
            {
                SPSecurity.RunWithElevatedPrivileges(delegate()
                {
                    getDataByListsURLs(lists, fields, _lastModified, nextItemID);
                });
            }
            else
            {
                getDataByListsURLs(lists, fields, _lastModified, nextItemID);
            }

            return result;
        }
        catch (Exception ex)
        {
            result.LoadXml(ex.Message);
            return result;
        }
    }
    
    #endregion WEBMETHODS

    #region METHODS

    private void InitializeComponent()
    {
        if (getWebConfigAppSettingsValue("SPMetadataRunWithElevatedPrivileges").ToLower() == "true")
        {
            runWithElevatedPrivileges = true;
        }

        if (getWebConfigAppSettingsValue("SPMetadataEnableLogging").ToLower() == "true")
        {
            verbose = true;
        }

        if (!string.IsNullOrEmpty(getWebConfigAppSettingsValue("SPMetadataLogFilePath")))
        {
            logFilePath = getWebConfigAppSettingsValue("SPMetadataLogFilePath");
        }

        if (getWebConfigAppSettingsValue("FixClaimsUserAccountRegex").ToLower() == "true")
        {
            fixClaimsUsername = true;
        }

        if (getWebConfigAppSettingsValue("SPMetadataItemsPerPage") != null)
        {
            if (!int.TryParse(getWebConfigAppSettingsValue("SPMetadataItemsPerPage"), out itemsPerPage))
            {
                itemsPerPage = 1000;
            }
        }

        if (getWebConfigAppSettingsValue("SPMetadataPagesPerBatch") != null)
        {
            if (!int.TryParse(getWebConfigAppSettingsValue("SPMetadataPagesPerBatch"), out pagesPerBatch))
            {
                pagesPerBatch = 10;
            }
        }

        currentPagingInfo = string.Empty;
    }

    private void getDataByListsURLs(string listsUrls, string fields, DateTime lastModified, int nextItemID)
    {
        string[] separators = { "|", "*" };
        string[] lists = listsUrls.Split(separators, StringSplitOptions.RemoveEmptyEntries);
        string innerXML = string.Empty;
        
        writeStartDocument();
        
        foreach (string list in lists)
        {
            try
            {
                innerXML = getItemsByListUrl(list, fields, lastModified, nextItemID);
            }
            catch
            { }

            writeListElement(list, currentPagingInfo, innerXML);
        }
    }

    private static string getItemsByListUrl(string ListURL, string fields, DateTime lastModified, int nextItemID)
    {
        string[] separators = { "|", "*" };
        string[] _fields = fields.Split(separators, StringSplitOptions.RemoveEmptyEntries);
        string result = string.Empty;
        
        using (SPSite _siteCollection = new SPSite(ListURL))
        {
            using (SPWeb _SPWeb = _siteCollection.OpenWeb())
            {
                SPList list = _SPWeb.GetListFromUrl(ListURL);
                if (nextItemID == 0)
                {
                    writeLog("Iterating items in list: '" + ListURL + "'. Total items: " + list.ItemCount);
                }
                try
                {
                    removeNonExistingFields(list, ref _fields);
                    if (_fields.Length > 0 && nextItemID == 0)
                    {
                        writeLog("Metadata fields to load: " + String.Join(",", _fields));
                    }
                    else if (_fields.Length <= 0)
                    {
                        throw new Exception("No metadata fields to load");
                    }
                    SPQuery oQuery = buildSPQuery2(_fields, list, lastModified, nextItemID);
                    return getItemsByListWithPaging(list, oQuery, _fields, lastModified);

                }
                catch (ArgumentException ex1)
                {
                    writeLog("Error: Argument Exception (Field does not exist): " + ex1.Message + ". StackTrace: " + ex1.StackTrace);
                }
                catch (SPQueryThrottledException ex2)
                {
                    writeLog("Error: SPQueryThrottledException: " + ex2.Message);
                }
                catch (Exception ex3)
                {
                    writeLog("Error: " + ex3.Message);
                }
            }
        }

        return string.Empty;
    }
    
    private static string getWebConfigAppSettingsValue(string key)
    {
        string val = string.Empty;
        string tempVal = System.Configuration.ConfigurationManager.AppSettings[key];
        if (!string.IsNullOrEmpty(tempVal)) val = tempVal;
        return val;
    }
    
    

    private static SPQuery buildSPQuery(string[] _fields, SPList oList, DateTime lastModified, int nextItemID)
    {
        SPQuery oQuery = new SPQuery();
        oQuery.RowLimit = (uint)itemsPerPage;// default 1000
        oQuery.QueryThrottleMode = SPQueryThrottleOption.Override;
        if (nextItemID > 0)
        {
            oQuery.ListItemCollectionPosition = new SPListItemCollectionPosition(@"Paged=TRUE&p_ID=" + nextItemID);
        }
        StringBuilder fieldsQuery = new StringBuilder();
        
        //fieldsQuery.Append(@"<ViewFields>"); cause error
        foreach (string field in _fields)
        {
            fieldsQuery.Append(string.Format("<FieldRef Name=\"{0}\" Nullable=\"True\" />", field.Replace(" ", @"_x0020_")));
        }
        //fieldsQuery.Append(@"</ViewFields>"); cause error
        if (fieldsQuery.Length == 0)
        {
            throw new Exception("no fields to load");
        }
        fieldsQuery.Append(string.Format("<Where><Geq><FieldRef Name='Modified'/><Value Type='DateTime'>{0}</Value></Geq></Where>", SPUtility.CreateISO8601DateTimeFromSystemDateTime(lastModified)));
        if (nextItemID == 0)
        {
            writeLog("CAML query: " + fieldsQuery.ToString());
        }
        oQuery.ViewFields = fieldsQuery.ToString();
        oQuery.ViewFieldsOnly = true;
        oQuery.Query = "";
        
        return oQuery;
    }

    private static SPQuery buildSPQuery2(string[] _fields, SPList oList, DateTime lastModified, int nextItemID)
    {
        StringBuilder fieldsQuery = new StringBuilder();
        SPQuery qry = new SPQuery();

        if (nextItemID > 0)
        {
            qry.ListItemCollectionPosition = new SPListItemCollectionPosition(@"Paged=TRUE&p_ID=" + nextItemID);
        }
        
        qry.Query =
        @"<Where><Geq><FieldRef Name='Modified' /><Value Type='DateTime'>" + SPUtility.CreateISO8601DateTimeFromSystemDateTime(lastModified) + @"</Value></Geq></Where><OrderBy><FieldRef Name='ID' /></OrderBy>";
        foreach (string field in _fields)
        {
            fieldsQuery.Append(string.Format("<FieldRef Name='{0}' />", field.Replace(" ", @"_x0020_")));
        }
        if (fieldsQuery.Length == 0)
        {
            throw new Exception("no fields to load");
        }
        qry.ViewFields = fieldsQuery.ToString();
        qry.RowLimit = (uint)itemsPerPage;
        
        return qry;
    }

    private static string getItemsByListWithPaging(SPList oList, SPQuery oQuery, string[] _fields, DateTime lastModified)
    {
        int currPageIndex = 1;
        int skippedItems = 0;
        StringBuilder res = new StringBuilder();
        
        do
        {
            writeLog("Current Page #" + currPageIndex);
            SPListItemCollection collListItems = oList.GetItems(oQuery);
            writeLog("Item(s) found: " + collListItems.Count);

            foreach (SPListItem oListItem in collListItems)
            {
                if (!isItemModified(oListItem, lastModified))
                {
                    ++skippedItems;
                    continue;
                }
                
                StringBuilder temp = new StringBuilder();
                Guid objectGuid = getSafeItemGuid(oList, oListItem);
                foreach (string field in _fields)
                {
                    string property = string.Empty;
                    if (TryGetProperty(oListItem, field, out property))
                    {
                        temp.Append("<" + field.Replace(' ', '_') + ">" + escapeXMLValue(property) + @"</" + field.Replace(' ', '_') + ">");
                    }
                }
                if (temp.Length > 0)
                {
                    res.Append(@"<item><spid>" + objectGuid.ToString() + @"</spid>" + temp + @"</item>");
                }
            }

            if (currPageIndex >= pagesPerBatch) // done with this batch, return
            {
                if (collListItems.Count < itemsPerPage || collListItems.ListItemCollectionPosition == null) //no more data available
                {
                    currentPagingInfo = string.Empty;
                }
                else // more data available
                {
                    // store current position
                    currentPagingInfo = collListItems.ListItemCollectionPosition.PagingInfo;
                }

                oQuery.ListItemCollectionPosition = null;
            }
            else
            {
                oQuery.ListItemCollectionPosition = collListItems.ListItemCollectionPosition;
            }
            
            currPageIndex++;
            
            if (skippedItems > 0)
            {
                writeLog("Item(s) skipped: " + skippedItems);
                skippedItems = 0;
            }
        } while (oQuery.ListItemCollectionPosition != null);

        return res.ToString();
    }

    private static bool isItemModified(SPListItem oListItem, DateTime lastModified)
    {
        try
        {
            if (oListItem["Modified"] != null && DateTime.Parse(oListItem["Modified"].ToString()) > lastModified)
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        catch(Exception ex)
        {
            writeLog("Could not get item last modification date. " + ex.Message);
            return true;
        }
    }
    
    private static bool TryGetProperty(SPListItem oListItem, string field, out string prop)
    {
        prop = string.Empty;

        try
        {

            try //try with space
            {
                if (oListItem[field] != null)
                {
                    prop = oListItem[field].ToString();
                }
            }
            catch (Exception ex)
            {
                if (ex.Message.Contains(@"Column '" + field + "' does not exist. It may have been deleted by another user."))
                {
                    field = field.Replace(" ", @"_x0020_");
                }
                else
                {
                    writeLog("Exception: " + ex.Message);
                }
            }
                
            if (oListItem[field] != null)
            {
                prop = oListItem[field].ToString();
            }
            if (!string.IsNullOrEmpty(prop))
            {
                if (prop.Contains('#'))
                {
                    prop = prop.Split('#')[1];
                }
                if (!prop.Contains("\\"))
                {
                    if (prop.Contains('|'))
                    {
                        prop = prop.Split('|')[0];
                    }
                }
                else //username
                {
                    prop = resolveUserName(prop);
                }
            }
            if (!string.IsNullOrEmpty(prop))
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        catch(Exception ex)
        {
            if (!ex.Message.Contains("Value does not fall within the expected range"))
            {
                writeLog("Could not get field '" + field + "' for item: '" + oListItem.Url + "'. " + ex.Message);
            }
            return false;
        }
    }
    
    private static Guid getSafeItemGuid(SPList oList, SPListItem oListItem)
    {
        return oList.GetItemById(oListItem.ID).UniqueId;
    }
    
    private static string escapeXMLValue(string value)
    {
        if (string.IsNullOrEmpty(value)) return value;
        string temp = value;
        temp = temp.Replace("&", "&amp;").Replace("<", "&lt;").Replace(">", "&gt;").Replace("\"", "&quot;").Replace("'", "&apos;");
        return temp;
    }

    private static string resolveUserName(string username)
    {
        //fix claims account name
        if(fixClaimsUsername)
        {
            Match res = Regex.Match(username, getWebConfigAppSettingsValue("FixClaimsUserAccountRegex"), RegexOptions.IgnoreCase | RegexOptions.IgnorePatternWhitespace);
            if (res.Success)
            {
                username = getUserFromRegex(res);
            }
        }

        return username;
    }

    private static string getUserFromRegex(Match matchResult)
    {
        if (string.IsNullOrEmpty(matchResult.Groups[2].Value))
        {
            return matchResult.Value;
        }
        else
        {
            return (matchResult.Groups[2].Value);
        }
    }

    private static void writeLog(string message)
    {
        if (verbose)
        {
            try
            {
                using (System.IO.StreamWriter sw = System.IO.File.AppendText(logFilePath.Replace(@"\", @"\\")))
                {
                    sw.Write(Environment.NewLine + DateTime.Now + "\t" + message);
                }
            }
            catch
            { }
        }
    }

    private static void removeNonExistingFields(SPList list, ref string[] _fields)
    {
        if (list != null && _fields.Length > 0)
        {
            SPFieldCollection spFields = list.Fields;
            
            foreach (string field in _fields)
            {
                if (!spFields.ContainsField(field))
                {
                    writeLog("Could not find field '" + field + "' in list '" + list.Title + "', Ignoring.");
                    _fields = _fields.Where(val => val != field).ToArray();
                }
            }
        }
    }

    private static int resolveItemIdFromPagingInfo(string pagingInfo)
    {
        //Paged=TRUE&p_ID=20
        int nextItemID = -1;

        if (!int.TryParse(pagingInfo.Substring(pagingInfo.IndexOf("p_ID") + 5), out nextItemID))
        {
            nextItemID = -1;
        }

        return nextItemID;
    }

    private void writeStartDocument()
    {
        XmlNode docNode = result.CreateXmlDeclaration("1.0", "UTF-8", null);
        result.AppendChild(docNode);
        XmlNode treeNode = result.CreateElement("tree");
        result.AppendChild(treeNode);
    }

    private void writeListElement(string list, string currentPagingInfo, string innerXML)
    {
        XmlNode listNode = result.CreateElement("list");
        XmlAttribute urlAttribute = result.CreateAttribute("url");
        urlAttribute.Value = list;
        listNode.Attributes.Append(urlAttribute);
        XmlAttribute hasMoreDataAttribute = result.CreateAttribute("hasmoredata");
        hasMoreDataAttribute.Value = (!string.IsNullOrEmpty(currentPagingInfo)) ? "true" : "false";
        listNode.Attributes.Append(hasMoreDataAttribute);
        if (!string.IsNullOrEmpty(currentPagingInfo))
        {
            XmlAttribute nextItemIdAttribute = result.CreateAttribute("nextitemid");
            nextItemIdAttribute.Value = resolveItemIdFromPagingInfo(currentPagingInfo).ToString();
            listNode.Attributes.Append(nextItemIdAttribute);
        }

        listNode.InnerXml = innerXML;
        listNode.Attributes.RemoveNamedItem("xmlns");
        result.SelectSingleNode(@"/tree").AppendChild(listNode);
    }
    
    #endregion METHODS
}
