﻿<%@ WebService Language="C#" Class="SpSecurity" %>

using System;
using System.Web;
using System.Web.Services;
using System.Collections.Generic;
using System.Text;
using System.Collections.Specialized;
using Microsoft.SharePoint;
using System.Configuration;
using Microsoft.SharePoint.Deployment;
using Microsoft.SharePoint.Administration;
using Microsoft.SharePoint.Utilities;
/// <summary>
/// Summary description for SpSecurity
/// </summary>
[WebService(Namespace = "http://tempuri.org/")]
[WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
// To allow this Web Service to be called from script, using ASP.NET AJAX, uncomment the following line. 
// [System.Web.Script.Services.ScriptService]
public class SpSecurity : System.Web.Services.WebService
{
    public SpSecurity()
    {
        //Uncomment the following line if using designed components 
        //InitializeComponent(); 
    }

    [WebMethod]
    public bool HasViewPermissions(string username, string url)
    {
        return SpSecurityHelper.HasViewPermissions(username, url);
    }
}

public static class SpSecurityHelper
{

    /// <summary>
    /// 
    /// </summary>
    /// <param name="username"></param>
    /// <param name="url"></param>
    /// <returns></returns>
    public static bool HasViewPermissions(string username, string url)
    {
        bool winAccess = false, claimsAccess = false, access = false;
        bool SpSecurityClaimsAuthentication = false;
        bool verbose = false;
        NameValueCollection queryParams = null;
        string claimsAuthPrefix = null;
        string claimsUserName = username, winUserName = username;

        try
        {
            try
            {
                SPSecurity.RunWithElevatedPrivileges(delegate()
                {
                    checkSPViewPermission(ref url, ref winAccess, ref claimsAccess, ref access, ref SpSecurityClaimsAuthentication, ref verbose, ref queryParams, ref claimsAuthPrefix, claimsUserName, ref winUserName);
                });
            }
            catch
            {
                checkSPViewPermission(ref url, ref winAccess, ref claimsAccess, ref access, ref SpSecurityClaimsAuthentication, ref verbose, ref queryParams, ref claimsAuthPrefix, claimsUserName, ref winUserName);
            }
        }
        catch (Exception ex)
        {
            writeLog(ex.Message + "\n" + ex.StackTrace);
            access = false;
            throw;
        }

        return access;
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="url"></param>
    /// <param name="winAccess"></param>
    /// <param name="claimsAccess"></param>
    /// <param name="access"></param>
    /// <param name="SpSecurityClaimsAuthentication"></param>
    /// <param name="verbose"></param>
    /// <param name="queryParams"></param>
    /// <param name="claimsAuthPrefix"></param>
    /// <param name="claimsUserName"></param>
    /// <returns></returns>
    private static void checkSPViewPermission(ref string url, ref bool winAccess, ref bool claimsAccess, ref bool access, ref bool SpSecurityClaimsAuthentication, ref bool verbose, ref NameValueCollection queryParams, ref string claimsAuthPrefix, string claimsUserName, ref string winUserName)
    {
        if (!string.IsNullOrEmpty(ConfigurationManager.AppSettings["SpSecurityClaimsAuthentication"])
            && ConfigurationManager.AppSettings["SpSecurityClaimsAuthentication"].ToLower() == "true")
        {
            SpSecurityClaimsAuthentication = true;
            claimsAuthPrefix = ConfigurationManager.AppSettings["ClaimsAuthenticationPrefix"];
        }

        if (!string.IsNullOrEmpty(ConfigurationManager.AppSettings["SpSecurityEnableLogging"])
            && ConfigurationManager.AppSettings["SpSecurityEnableLogging"].ToLower() == "true")
            verbose = true;

        url = HttpUtility.UrlDecode(url);

        if (SpSecurityClaimsAuthentication && !string.IsNullOrEmpty(claimsAuthPrefix)) //using claims auth.
        {
            winUserName = winUserName.Remove(0, claimsAuthPrefix.Length);
        }

        winAccess = DoesUserHaveViewPermission(winUserName, url, verbose);
        if (!winAccess && SpSecurityClaimsAuthentication)
            claimsAccess = DoesUserHaveViewPermission(claimsUserName, url, verbose);
        access = winAccess || claimsAccess;
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="username"></param>
    /// <param name="url"></param>
    /// <param name="verbose"></param>
    /// <returns></returns>
    private static bool DoesUserHaveViewPermission(string username, string url, bool verbose)
    {
        bool access = false;
        
        try
        {
            try
            {
                SPSecurity.RunWithElevatedPrivileges(delegate()
                {
                    checkSPUserPermission(username, url, verbose, out access);
                    return;
                });
            }
            catch
            {
                checkSPUserPermission(username, url, verbose, out access);
            }
        }
        catch
        {
            access = false;
        }

        return access;
    }
    
    /// <summary>
    /// 
    /// </summary>
    /// <param name="username"></param>
    /// <param name="url"></param>
    /// <param name="verbose"></param>
    /// <param name="access"></param>
    private static void checkSPUserPermission(string username, string url, bool verbose, out bool access)
    {
        SPSite tempSite = null;
        SPWeb tempWeb = null;
        SPList list = null;
        SPListItem listItem = null;
        NameValueCollection queryParams = null;
        SPDeploymentObjectType objType = SPDeploymentObjectType.Invalid;
        SPBasePermissions permsToCheck= SPBasePermissions.EmptyMask;
        bool _continue = false;
        string permissionLevelToCheck = null;
        string defaultPage = null;
        string defaultRole = "Full Control";
        
        access = false;

        try
        {
            permissionLevelToCheck = getPermissionLevel(defaultRole, verbose);
            if (verbose) writeLog("SpSecurity: Verifying '" + username + "' has '" + permissionLevelToCheck + "' permission on URL: '" + url + "'...");
            
            // if user got permissions from Web Application User Policy
            if (CheckWebAppPolicies() &&  getUserPermissionsFromUserPolicy(username, url, verbose).Contains(permissionLevelToCheck))
            {
                if (verbose) writeLog("SpSecurity: '" + username + "' has permissions through User Policy, access granted.");
                access = true;
                return;
            }
            
            tempSite = new SPSite(url);
            tempWeb = tempSite.OpenWeb();
            Guid objID = Guid.Empty;

            // check if entire web is anonymous
            if (tempWeb.AnonymousState == SPWeb.WebAnonymousState.On)
            {
                if (verbose) writeLog("SpSecurity: Web site is anonymous, access granted.");
                access = true;
                return;
            }

            // check welcome page
            defaultPage = tempWeb.RootFolder.WelcomePage;
            if (string.IsNullOrEmpty(defaultPage)) defaultPage = "default.aspx";

            // get SharePoint object type from Url
            queryParams = getQueryParams(url);
            objType = GetObjectTypeFromURL(url, queryParams, out objID);
            if (objType.Equals(SPDeploymentObjectType.Invalid) && verbose) 
                writeLog("SpSecurity: Could not determine SharePoint object type from url.");

            // create impersonated user token
            SPUser oUser = getSPUserFromSPWeb(tempWeb, username);
            SPUserToken usertoken = tempWeb.AllUsers[oUser.LoginName].UserToken;

            // check site collection admin
            checkSiteCollectionAdmin(url, oUser, verbose, out _continue, ref access);
            if (access)
                return;
            //if (!_continue)
            //    return;

            // check permission for the SharePoint object of the url
            using (SPSite oSiteCollection = new SPSite(url))
            {
                using (SPWeb oWebsite = oSiteCollection.OpenWeb())
                {
                    permsToCheck = getSPBasePermissionsToCheck(oWebsite, permissionLevelToCheck);
                    bool previousState_CatchAccessDeniedException = oSiteCollection.CatchAccessDeniedException;
                    
                    oSiteCollection.CatchAccessDeniedException = false; // not caught and handled by Microsoft SharePoint Foundation 2010
                    
                    // if welcome page check view permission on the web level
                    if(isWelcomePage(oWebsite,defaultPage,url,verbose))
                    {
                        if (verbose) writeLog("SpSecurity: Url is a welcome page. Checking permission on web level...");
                        access = oWebsite.DoesUserHavePermissions(username, permsToCheck);
                        //logCurrentUserPerms(oUser, usertoken, url);
                    }
                    else // url is not welcome page, check all other objects
                    {
                        if (verbose && objType != SPDeploymentObjectType.Invalid) writeLog("SpSecurity: Identified SharePoint object type '" + objType.ToString() + "'");
                        switch (objType)
                        {
                            case SPDeploymentObjectType.ListItem:
                                {
                                    //if (verbose) writeLog("SpSecurity: Identified SharePoint object 'SPListItem'");
                                    list = oWebsite.GetListFromUrl(url);
                                    if (list != null)
                                        listItem = list.Items[objID];
                                    if (listItem != null)
                                        access = listItem.DoesUserHavePermissions(oUser, permsToCheck);
                                }
                                break;
                            case SPDeploymentObjectType.List:
                                {
                                    list = oWebsite.Lists[objID];
                                    if (list != null)
                                        access = list.DoesUserHavePermissions(oUser, permsToCheck);
                                }
                                break;
                            case SPDeploymentObjectType.Folder:
                                {
                                    SPFolder folder = oWebsite.GetFolder(objID);
                                    if (folder != null)
                                        if (folder.Item != null)
                                            access = folder.Item.DoesUserHavePermissions(oUser, permsToCheck);
                                }
                                break;
                            case SPDeploymentObjectType.File:
                                {
                                    access = checkSPFileWithUserContext(usertoken, url, objID, permsToCheck);
                                }
                                break;
                            default:
                                {
                                    if (verbose) writeLog("SpSecurity: Invalid SharePoint object type.");
                                    access = false;
                                    break;
                                }

                        }
                    }
                    oSiteCollection.CatchAccessDeniedException = previousState_CatchAccessDeniedException;
                    if (verbose)
                    {
                        if (access) writeLog("SpSecurity: '" + username + "' has permission, access granted.");
                        else writeLog("SpSecurity: '" + username + "' doesn't have permission, access denied.");
                    }
                }
            }
        }
        catch (Exception ex)
        {
            if (verbose) writeLog("SpSecurity: Could not check user permissions, access denied. Exception: " + ex.Message);
        }
        finally
        {
            if (tempWeb != null)
            {
                tempWeb.Dispose();
            }
            if (tempSite != null)
            {
                tempSite.Dispose();
            }
        }
    }


    //private static void logCurrentUserPerms(SPUser user, SPUserToken usertoken, string url)
    //{
    //    using(SPSite site = new SPSite(url, usertoken))
    //    {
    //        using(SPWeb web = site.OpenWeb())
    //        {
    //            foreach(var role in web.AllRolesForCurrentUser)
    //            {
    //            }
    //        }
    //    }
        
    //    //SPRoleAssignment assignment = web.RoleAssignments.GetAssignmentByPrincipal((SPPrincipal)user);

    //    //StringBuilder sb = new StringBuilder("");

    //    //foreach (SPRoleDefinition role in assignment.RoleDefinitionBindings)
    //    //{

    //    //    String _role = role.Name;

    //    //    String permission = role.BasePermissions.ToString();

    //    //    writeLog("Role: " + permission);

    //    //}

    //    //Console.WriteLine(“\n\n Roles Assignments:”);

    //    //foreach (SPRoleAssignment roleA in web.RoleAssignments)
    //    //{
    //    //    //Console.WriteLine(“The following Role definition bindings exist for ” +
    //    //    //roleA.Member.Name);
    //    //    foreach (SPRoleDefinition roledef in roleA.RoleDefinitionBindings)
    //    //    {
    //    //        //Console.WriteLine(roledef.Name);
    //    //    }
    //    //}
    //    //Console.ReadLine();

    //}

    
    /// <summary>
    /// 
    /// </summary>
    /// <param name="url"></param>
    /// <param name="queryParams"></param>
    /// <returns></returns>
    private static NameValueCollection getQueryParams(string url)
    {
        string querystring = null;
        int index = url.IndexOf("?");

        NameValueCollection queryParams = null;

        if (index != -1)
        {
            querystring = url.Substring(index + 1);
            queryParams = HttpUtility.ParseQueryString(querystring);
        }
        
        return queryParams;
    }

    private static string getBaseUrl(string url)
    {
        string baseUrl = url;
        int index = url.IndexOf("?");

        if (url.EndsWith("/"))
            baseUrl = url.TrimEnd('/');

        if (index != -1)
        {
            baseUrl = url.Substring(0, index);
        }
        
        return baseUrl;
    }
    
    /// <summary>
    /// This method tries to check if a given user has view permission over a given SPFile, it tries to do so in three ways
    /// </summary>
    /// <param name="userToken"></param>
    /// <param name="Url"></param>
    /// <returns></returns>
    private static bool checkSPFileWithUserContext(object userTokenObject, string Url, Guid objId, SPBasePermissions permToCheck)
    {
        SPUserToken userToken = userTokenObject as SPUserToken;
        SPSite site = new SPSite(Url, userToken);
        SPWeb web = site.OpenWeb();
        object obj = null;
        bool hasAccess = false;
        bool catchAccessDeniedException = site.CatchAccessDeniedException;

        site.CatchAccessDeniedException = false;  //this exception is not caught in sp 2010
        
        // first try
        try
        {
            obj = web.GetFileOrFolderObject(Url);
            if (obj is SPFile)
            {
                SPFile f = (SPFile)obj;
                if (f.Item != null)
                {
                    hasAccess = f.Item.DoesUserHavePermissions(permToCheck);
                    if (hasAccess)
                        return true;
                }
            }
        }
        catch { }
        // second try
        try
        {
            obj = web.GetFile(objId);
            if (obj is SPFile)
            {
                SPFile f = (SPFile)obj;
                if (f.Item != null)
                {
                    hasAccess = f.Item.DoesUserHavePermissions(permToCheck);
                    if (hasAccess)
                        return true;
                }
            }
        }
        catch { }
        // third try
        try
        {
            if (obj is SPFile)
            {
                SPFile f = (SPFile)obj;
                SPFile f2 = web.Files[f.Url];
                if (f2.Item != null)
                {
                    hasAccess = f2.Item.DoesUserHavePermissions(permToCheck);
                    if (hasAccess)
                        return true;
                }
            }
        }
        catch (Exception ex)
        {
            throw ex;
        }
            
        finally
        {
            site.CatchAccessDeniedException = catchAccessDeniedException;
            if (web != null)
            {
                web.Dispose();
            }
            if (site != null)
            {
                site.Dispose();
            }
        }
        return hasAccess;
    }

    /// <summary>
    /// This method returns the SharePoint object type (Site/Web/List/ListItem/Folder/File) of a given url. If none found, SPDeploymentObjectType.Invalid is returned.
    /// </summary>
    /// <param name="baseUrl"></param>
    /// <param name="queryParams"></param>
    /// <param name="objectGuid"></param>
    /// <returns>SPDeploymentObjectType</returns>
    private static SPDeploymentObjectType GetObjectTypeFromURL(string baseUrl, NameValueCollection queryParams, out Guid objectGuid)
    {
        SPDeploymentObjectType objType = SPDeploymentObjectType.Invalid;
        SPSite site = null;
        SPWeb web = null;
        SPList list = null;
        SPFolder folder = null;
        objectGuid = Guid.Empty;
        object obj = null;
        string idValue = null;
        string folderValue = null;
        string contentTypeId = null;
        
        try
        {
            site = new SPSite(baseUrl);
            web = site.OpenWeb();
            
            if (queryParams != null)
            {
                idValue = queryParams["ID"];
                folderValue = queryParams["RootFolder"];
                contentTypeId = getContentTypeIdFromQueryParams(queryParams);
            }
            if (idValue == null && folderValue == null)
            {
                // check SPSite
                if (baseUrl.Equals(site.Url, StringComparison.OrdinalIgnoreCase))
                {
                    objectGuid = site.ID;
                    return SPDeploymentObjectType.Site;
                }

                // check SPWeb
                if (baseUrl.Equals(web.Url, StringComparison.OrdinalIgnoreCase))
                {
                    objectGuid = web.ID;
                    return SPDeploymentObjectType.Web;
                }

                // check SPList
                string siteRelativeUrl = web.ServerRelativeUrl;
                list = web.GetList(baseUrl);
                
                if ((list != null) && (siteRelativeUrl.Equals(list.ParentWebUrl, StringComparison.OrdinalIgnoreCase)))   
                {
                    objectGuid = list.ID;
                    return SPDeploymentObjectType.List;
                }
            }
            
            // check SPFile/SPFolder/SPListItem
            obj = web.GetObject(baseUrl);
            if (list == null)
                list = web.GetList(baseUrl);
            if (obj != null)
            {
                if (obj is SPFile)
                {
                    SPFile file = obj as SPFile;
                    if (queryParams != null)
                    {
                        if (idValue != null && contentTypeId == null) //list item sp2007
                        {
                            SPListItem item = list.GetItemById(int.Parse(idValue));
                            objectGuid = item.UniqueId;
                            return SPDeploymentObjectType.ListItem;
                        }
                        else if (idValue != null && contentTypeId != null) //list item sp2013
                        {
                            if (!contentTypeId.Contains(@"0x0120"))
                            {
                                SPListItem item = list.GetItemById(int.Parse(idValue));
                                objectGuid = item.UniqueId;
                                return SPDeploymentObjectType.ListItem;
                            }
                        }
                        if (folderValue != null && contentTypeId != null) //folder
                        {
                            if (contentTypeId.Contains(@"0x0120"))
                            {
                                if (baseUrl.Contains(@"AllItems.aspx?RootFolder=")) // for urls like: "http://website/Lists/links/AllItems.aspx?RootFolder=/Lists/links/foldername&FolderCTID=0x0120005452CEC5E473E148AF28230C75906BC8"
                                {
                                    folder = web.GetFolder(folderValue);
                                    objectGuid = folder.UniqueId;
                                }
                                else if (idValue != null)// for urls like: "http://website/Lists/links/DispForm.aspx?ID=2&Source=http://website/Lists/links/AllItems.aspx&ContentTypeId=0x0120005452CEC5E473E148AF28230C75906BC8&RootFolder=/Lists/links"
                                {
                                    foreach (SPListItem fold in list.Folders)
                                    {
                                        if (fold.ID == int.Parse(idValue))
                                            objectGuid = fold.UniqueId;

                                    }
                                }
                                return SPDeploymentObjectType.Folder;
                            }
                        }
                        else
                        {
                        }
                    }
                    else
                    {
                        objectGuid = file.UniqueId;
                        return SPDeploymentObjectType.File;
                    }
                }
                else if (obj is SPFolder)
                {
                    folder = obj as SPFolder;
                    objectGuid = folder.UniqueId;
                    return SPDeploymentObjectType.Folder;
                }
                else if (obj is SPListItem)
                {

                    SPListItem item = obj as SPListItem;

                    if (null != item.Folder)
                    {
                        objectGuid = item.Folder.UniqueId;
                        return SPDeploymentObjectType.Folder;
                    }
                    else if (null != item.File)
                    {
                        objectGuid = item.File.UniqueId;
                        return SPDeploymentObjectType.File;
                    }

                    objectGuid = item.UniqueId;
                    return SPDeploymentObjectType.ListItem;
                }
            }
        }
        finally
        {
            if (web != null)
            {
                web.Dispose();
            }
            if (site != null)
            {
                site.Dispose();
            }
        }
        return objType;
    }

    private static string removeEndingPartFromUrl(string url)
    {
        return (url.Substring(0, url.LastIndexOf(@"/")));
    }

    private static bool isSiteCollectionAdmin(string url, SPUser oUser, bool verbose)
    {
        try
        {
            using (SPSite oSiteCollection = new SPSite(url))
            {
                using (SPWeb oWebsite = oSiteCollection.OpenWeb())
                {
                    if (oUser.IsSiteAdmin)
                    {
                        return true;
                    }
                }
            }
            return false;
        }
        catch (Exception ex)
        {
            if (verbose) writeLog("SpSecurity: Could not determine if '" + oUser.LoginName + "' is a site collection administrator. Exception: " + ex.Message);
            return false;
        }
    }

    private static bool allowSiteCollectionAdminOnly(bool verbose)
    {
        if (!string.IsNullOrEmpty(ConfigurationManager.AppSettings["SpSecurityEnableSiteStatisticsForSiteAdminsOnly"])
            && ConfigurationManager.AppSettings["SpSecurityEnableSiteStatisticsForSiteAdminsOnly"].ToLower() == "true")
        {
            //if (verbose) writeLog("SpSecurity: Site Collection Administrator privileges are required.");
            return true;
        }
        return false;
    }

    private static string getPermissionLevel(string defaultRole, bool verbose)
    {
        string permLevel = string.Empty;
        string[] defaultPermissionLevels = {
                                      // SP 2007 + SP 2010:
                                        "Full Control", "Design", "Contribute", "Read", "Limited Access",
                                      // SP 2013:
                                        "View Only", "Edit", "Approve", "Manage Hierarchy", "Restricted Read", "Restricted Interfaces for Translation" };

        if (!string.IsNullOrEmpty(ConfigurationManager.AppSettings["SpSecurityPermissionLevel"]))
        {
            permLevel = ConfigurationManager.AppSettings["SpSecurityPermissionLevel"];
            int pos = Array.IndexOf(defaultPermissionLevels, permLevel);
            if (pos > -1)
            {
                return permLevel;
            }
            else //bad input
            {
                throw new Exception("Could not find value of '" + permLevel + "' in SharePoint default permission levels");
            }
        }
        else // empty or null, use default 'Full Control'
        {
            permLevel = defaultRole;
        }

        return permLevel;
    }
    
    private static void writeLog(string message)
    {
        try
        {
            using (System.IO.StreamWriter sw = System.IO.File.AppendText(ConfigurationManager.AppSettings["SpSecurityLogFilePath"].Replace(@"\", @"\\")))
            {
                sw.Write(Environment.NewLine + DateTime.Now + "\t" + message);
            }
        }
        catch (Exception)
        { }
    }

    /// <summary>
    /// run custom check according to special requirements.
    /// </summary>
    /// <param name="url"></param>
    /// <param name="username"></param>
    /// <param name="oUser"></param>
    /// <param name="verbose"></param>
    /// <param name="_continue">update this param to allow continuation of the security check</param>
    /// <param name="access"></param>
    private static void checkSiteCollectionAdmin(string url, SPUser oUser, bool verbose, out bool _continue, ref bool access)
    {
        try
        {
            if (allowSiteCollectionAdminOnly(verbose))
            {
                if (isSiteCollectionAdmin(url, oUser, verbose))
                {
                    access = true;
                    if (verbose) writeLog("SpSecurity: '" + oUser.LoginName + "' is a site collection administrator.");
                }

                else
                {
                    access = false;
                    if (verbose) writeLog("SpSecurity: '" + oUser.LoginName + "' is not a site collection administrator.");
                }
                _continue = false;
                return;
            }
            _continue = true;
        }
        catch (Exception ex)
        {
            access = false;
            _continue = false;
            if (verbose) writeLog("SpSecurity: Could not check if '" + oUser.LoginName + "' is a site collection administrator. Exception: " + ex.Message);
        }
            
    }

    private static bool isWelcomePage(SPWeb web, string defaultPage, string url, bool verbose)
    {
        if (url.Equals(web.Url + "/" + defaultPage, StringComparison.OrdinalIgnoreCase)
                            || url.Equals(web.Url, StringComparison.OrdinalIgnoreCase))
        {
            return true;
        }
        return false;
    }

    private static SPBasePermissions getSPBasePermissionsToCheck(SPWeb web, string roleToCheck)
    {
        SPBasePermissions permsToCheck = SPBasePermissions.EmptyMask;

        try
        {
            if (web.RoleDefinitions[roleToCheck] != null)
            {
                permsToCheck = web.RoleDefinitions[roleToCheck].BasePermissions;
            }
            else
            {
                throw new Exception("Unable to get role definition");
            }
            
            return permsToCheck;
        }
        catch (Exception ex)
        {
            writeLog("SpSecurity: Could not get required permissions to check, access denied. Exception: " + ex.Message);
            throw ex;
        }
    }

    private static string getContentTypeIdFromQueryParams(NameValueCollection queryParams)
    {
        string CTID = string.Empty;

        CTID = string.IsNullOrEmpty(queryParams["ContentTypeId"]) ? queryParams["CTID"] : queryParams["ContentTypeId"];
        CTID = string.IsNullOrEmpty(CTID) ? queryParams["FolderCTID"] : string.Empty;
        
        return CTID;
    }
    
    private static string getDomainFromURL(string url)
    {
        try
        {
            Uri uri = new Uri(url);
            return uri.Scheme + @"://" + uri.Host + ":" + uri.Port;
        }
        catch (Exception ex)
        {
            writeLog("SpSecurity: Could not resolve domain. Exception: " + ex.Message);
            return string.Empty;
        }
    }

    private static string getUserPermissionsFromUserPolicy(string user, string url, bool verbose)
    {
        string result = string.Empty;
        bool reachedMaxCount;
        SPPrincipalInfo pi = null;
        SPPolicyCollection policies = null;
        SPWebApplication oWebApplication = null;
        SPSite oSiteCollection = null;
        SPWeb oWebsite = null;
        
        try
        {
            if (verbose) writeLog("SpSecurity: Getting web application url for domain: '" + getDomainFromURL(url) + "'...");
            oWebApplication = SPWebApplication.Lookup(new Uri(getDomainFromURL(url)));
            if (verbose) writeLog("SpSecurity: Getting policies for web application: '" + oWebApplication.GetResponseUri(SPUrlZone.Default) + "'...");
            if (oWebApplication != null)
                policies = oWebApplication.Policies;
            else
                throw new Exception("Could not retrieve SPWebApplication object");
            
			if (verbose) writeLog("SpSecurity: Found " + policies.Count + " policies.");
            
            // go through users
            if (verbose) writeLog("SpSecurity: iterating user policies...");
            foreach (SPPolicy policy in policies)
            {
                if (verbose) writeLog("SpSecurity: Current policy: " + policy.DisplayName);
                try
                {
                    pi = SPUtility.ResolvePrincipal(oWebApplication, null, policy.UserName, SPPrincipalType.All, SPPrincipalSource.All, false);
                    if (pi != null)
                    {
                        if (pi.PrincipalType == SPPrincipalType.User)
                        {
                            if (policy.UserName.Equals(user, StringComparison.OrdinalIgnoreCase))
                            {
                                result = policy.PolicyRoleBindings.ToString();
                                if (verbose) writeLog("SpSecurity: User found, Permission Policy level: '" + policy.PolicyRoleBindings.ToString() + "'");
                                return result;
                            }
                        }
                    }
                    else
                    {
                        writeLog("Could not resolve principal for policy: " + policy.DisplayName);
                    }
                }
                catch(Exception ex)
                {
                    if (verbose) writeLog("SpSecurity: Could not check policy '" + policy.DisplayName + "', continuing to next policy. Exception: " + ex.Message);
                    //continue to next policy
                }
            }
            
            // go through groups
            if (verbose) writeLog("SpSecurity: iterating group policies...");
            foreach (SPPolicy policy in policies)
            {
                if (verbose) writeLog("SpSecurity: Current policy: " + policy.DisplayName);
                try
                {
                    pi = SPUtility.ResolvePrincipal(oWebApplication, null, policy.UserName, SPPrincipalType.All, SPPrincipalSource.All, false);
                    if (pi != null)
                    {
                        if (pi.PrincipalType == SPPrincipalType.SecurityGroup)
                        {
                            using (oSiteCollection = new SPSite(url))
                            {
                                using (oWebsite = oSiteCollection.OpenWeb())
                                {
                                    SPPrincipalInfo[] adUsers = SPUtility.GetPrincipalsInGroup(oWebsite, policy.UserName, int.MaxValue - 1, out reachedMaxCount);
                                    foreach (SPPrincipalInfo adUser in adUsers)
                                    {
                                        if (adUser.LoginName.Equals(user, StringComparison.OrdinalIgnoreCase))
                                        {
                                            result = policy.PolicyRoleBindings.ToString();
                                            if (verbose) writeLog("SpSecurity: User found in group '" + policy.UserName + "'. Group Permission Policy level: '" + policy.PolicyRoleBindings.ToString() + "'");
                                            return result;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        writeLog("Could not resolve principal for policy: " + policy.DisplayName);
                    }
                }
                catch (Exception ex)
                {
                    if (verbose) writeLog("SpSecurity: Could not check policy '" + policy.DisplayName + "', continuing to next policy. Exception: " + ex.Message);
                    //continue to next policy
                }
            }
        }
        catch (Exception ex)
        {
            if (verbose) writeLog("SpSecurity: Could not get Permission Policy. Exception: " + ex.Message);
        }
        finally
        {
            if (oWebsite != null)
            {
                oWebsite.Dispose();
            }
            if (oSiteCollection != null)
            {
                oSiteCollection.Dispose();
            }
        } 
        return result;
    }

    private static bool CheckWebAppPolicies()
    {
        string checkPolicies = ConfigurationManager.AppSettings["SpSecurityCheckWebAppPolicies"];
        if (!string.IsNullOrEmpty(checkPolicies) && checkPolicies.ToLower() == "true")
            return true;
        else
            return false;
    }

    private static SPUser getSPUserFromSPWeb(SPWeb web, string username)
    {
        SPUser oUser;
        try
        {
            oUser = web.EnsureUser(username);
            return oUser;
        }
        catch
        {
            //if we got here, that means the user either does not exist or its permissions were not given in the site
            writeLog("SpSecurity: Could not find '" + username + "' in website.");
            throw new Exception("User not found");
        }
    }
}

